from __future__ import (division, print_function)
from future import standard_library

from Utils.Utilities import decodeBytesToUnicode

standard_library.install_aliases()

import logging
from urllib.parse import urlparse

from collections import defaultdict
from WMCore.Services.Service import Service
from WMCore.Services.TagCollector.XMLUtils import xml_parser

class TagCollector(Service):
    """
    Class which provides interface to CMS TagCollector web-service.
    Provides non-deprecated CMSSW releases in all their ScramArchs (not only prod)
    """

    def __init__(self, url=None, logger=None, configDict=None, **kwargs):
        """
        responseType will be either xml or json
        """
        defaultURL = "https://cmssdt.cern.ch/SDT/cgi-bin/ReleasesXML"
        url = url or defaultURL
        parsedUrl = urlparse(url)
        self.cFileUrlPath = parsedUrl.path.replace("/", "_")
        # all releases types and all their archs
        self.tcArgs = kwargs
        self.tcArgs.setdefault("anytype", 1)
        self.tcArgs.setdefault("anyarch", 1)

        configDict = configDict or {}
        configDict.setdefault('endpoint', url)
        configDict.setdefault("timeout", 300)
        configDict.setdefault('cacheduration', 1)
        configDict['logger'] = logger if logger else logging.getLogger()
        super(TagCollector, self).__init__(configDict)
        self['logger'].debug("Initializing TagCollector with url: %s", self['endpoint'])

    def _getResult(self, callname="", clearCache=False,
                   args=None, verb="GET", encoder=None, decoder=None,
                   contentType=None):
        """
        _getResult_

        retrieve JSON/XML formatted information given the service name and the
        argument dictionaries

        TODO: Probably want to move this up into Service
        """
        if not args:
            args = self.tcArgs

        cFile = '%s_%s'% (self.cFileUrlPath, callname.replace("/", "_"))
        # If no callname or url path, the base host is getting queried
        if cFile == '_':
            cFile = 'baseRequest'

        if clearCache:
            self.clearCache(cFile, args, verb)

        # Note cFile is just the base name pattern, args 
        # are also considered for the end filename in the method below
        f = self.refreshCache(cFile, callname, args, encoder=encoder, decoder=decodeBytesToUnicode,
                              verb=verb, contentType=contentType)
        result = f.read()
        f.close()

        # overhead from REST model which returns results as strings or None
        # therefore they can be encoded by JSON to None, etc.
        if result == 'None':
            return
        if result and decoder:
            result = decoder(result)
        return result

    def data(self):
        "Fetch data from tag collector or local cache"
        data = self._getResult()
        pkey = 'architecture'
        for row in xml_parser(data, pkey):
            yield row[pkey]

    def releases(self, arch=None):
        "Yield CMS releases known in tag collector"
        arr = []
        for row in self.data():
            if arch:
                if arch == row['name']:
                    for item in row['project']:
                        arr.append(item['label'])
            else:
                for item in row['project']:
                    arr.append(item['label'])
        return list(set(arr))

    def architectures(self):
        "Yield CMS architectures known in tag collector"
        arr = []
        for row in self.data():
            arr.append(row['name'])
        return list(set(arr))

    def releases_by_architecture(self):
        "returns CMS architectures and realease in dictionary format"
        arch_dict = defaultdict(list)
        for row in self.data():
            releases = set()
            for item in row['project']:
                releases.add(item['label'])
            arch_dict[row['name']].extend(list(releases))
        return dict(arch_dict)


    def defaultMicroArchVersionNumberByRelease(self, default_microarch=0):
        """
        Yield default microarchitecture by CMS release known in tag collector.

        :param default_microarch: int, default microar when not found in TagCollector XML file.
        :return: dictionary ("release": default_microarch (or default_microarch if not found)).
        """
        rel_microarchs = {}
        for row in self.data():
            for item in row['project']:
                microarch = item.get('default_micro_arch', default_microarch)
                if isinstance(microarch, str):
                    # format of this value is "arch-vN"
                    microarch = int((microarch.split("-")[-1]).lstrip("v"))
                rel_microarchs[item['label']] = microarch
        return rel_microarchs

    def getGreaterMicroarchVersionNumber(self, releases=None, default_microarch=0, rel_microarchs=None):
        """
        Return the greated default microarchitecture, given a list of releases.

        :param releases: str, comma-separated list of releases.
        :param default_microarch: int, default microar when not found in TagCollector XML file.
        :return: int, greater default microarchitecture version number.
        """
        if releases is None:
            return default_microarch

        if rel_microarchs is None:
            rel_microarchs = self.defaultMicroArchVersionNumberByRelease()

        releases = releases.split(",")
        microarch = default_microarch
        for r in releases:
            m = rel_microarchs.get(r, default_microarch)
            if m > microarch:
                microarch = m

        return microarch



'''
Created on Jul 31, 2014

@author: sryu
'''
from __future__ import print_function, division
from builtins import object

import cherrypy
import traceback
from threading import Thread, Condition

from WMCore.WMLogging import getTimeRotatingLogger


class CherryPyPeriodicTask(object):

    def __init__(self, config):

        """
        BaseClass which can set up the concurrent task using cherrypy thread.
        WARNING: This assumes each task doesn't share the object.
        (It can be share only read operation is performed)
        If the object shared by multple task and read/write operation is performed.
        Lock is not provided for these object

        :arg config  WMCore.Configuration object. which need to contain in duration attr.
        TODO: add validation for config.duration
        """
        self.logger = getTimeRotatingLogger(config._internal_name, config.log_file)
        self.logger.info("Setting CherryPy periodic task with the following config:\n%s", config)
        self.setConcurrentTasks(config)
        self.setUpLogDB(config)

        for task in self.concurrentTasks:
            PeriodicWorker(task['func'], config, task['duration'], logger=self.logger, logDB=self.logDB)

    def setUpLogDB(self, config):
        if hasattr(config, "central_logdb_url"):
            # default set up for logDB config section need to contain propervalues
            from WMCore.Services.LogDB.LogDB import LogDB
            self.logDB = LogDB(config.central_logdb_url, config.log_reporter, logger=self.logger,
                               thread_name=config.object.rsplit(".", 1)[-1])
        else:
            self.logDB = None

    def setConcurrentTasks(self, config):
        """
        sets the list of function reference for concurrent tasks,
        sub class should implement this

        each function in the list should have the same signature with
        3 arguments (self, config, duration)
        config is WMCore.Configuration object
        """
        self.concurrentTasks = {'func': None, 'duration': None}
        raise NotImplementedError("need to implement setSequencialTas assign self._callSequence")

class PeriodicWorker(Thread):

    def __init__(self, func, config, duration=600, logger=cherrypy.log, logDB=None):
        # use default RLock from condition
        # Lock wan't be shared between the instance used  only for wait
        # func : function or callable object pointer
        self.wakeUp = Condition()
        self.stopFlag = False
        self.taskFunc = func
        self.config = config
        self.duration = duration
        self.logger = logger
        self.logDB = logDB

        try:
            name = func.__name__
            print(name)
        except AttributeError:
            name = func.__class__.__name__
            print(name)

        Thread.__init__(self, name=name)
        cherrypy.engine.subscribe('start', self.start, priority=100)
        cherrypy.engine.subscribe('stop', self.stop, priority=100)

    def stop(self):
        self.wakeUp.acquire()
        self.stopFlag = True
        self.wakeUp.notifyAll()
        self.wakeUp.release()

    def run(self):

        while not self.stopFlag:
            self.wakeUp.acquire()
            try:
                self.taskFunc(self.config)
                self.heartBeatInfoToLogDB()
            except Exception as e:
                self.logger.error("Periodic Thread ERROR %s.%s %s"
                % (getattr(e, "__module__", "__builtins__"),
                e.__class__.__name__, str(e)))
                traceMsg = traceback.format_exc()
                for line in traceMsg.rstrip().split("\n"):
                    self.logger.error(" " + line)
                self.heartBeatErrorToLogDB(traceMsg)

            self.wakeUp.wait(self.duration)
            self.wakeUp.release()

    def heartBeatInfoToLogDB(self):
        if self.logDB:
            self.logDB.delete(mtype="error", this_thread=True, agent=False)
            self.logDB.post(mtype="info")
        return

    def heartBeatErrorToLogDB(self, msg):
        if self.logDB:
            self.logDB.post(msg=msg, mtype="error")
        return


class SequentialTaskBase(object):

    """
    Base class for the tasks which should run sequentially
    """
    def __init__(self):
        self.setCallSequence()

    def __call__(self, config):
        for call in self._callSequence:
            try:
                call(config)
            except Exception as ex:
                #log the excpeiotn and break.
                #SequencialTasks are interconnected between functions
                trace = traceback.format_exc()
                self.logger.error("%s:\n %s" % (str(ex), trace))
                break

    def setCallSequence(self):
        """
        set the list of function call with out args on self.callSequence

        i.e.
        self.callSequence = [self.do_something1, self.do_something1]
        """
        raise NotImplementedError("need to implement setCallSequence assign self._callSequence")


#this is the sckeleton of request data collector
class DataUploadTask(SequentialTaskBase):

    def setCallSequence(self):
        self._callSequence = [self.getData, self.convertData, self.putData]

    def getData(self, config):
        # self.data = getData(self.sourceUrl)
        pass

    def convertData(self, config):
        # self.data = convertData(self.data)
        pass

    def putData(self, config):
        # putData(self.destUrl)
        pass



"""
File       : MSTransferorError.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSTransferorError represents MSTransferor errors
"""

# system modules
import json

# WMCore modules
from WMCore.WMException import WMException


# MSTransferor error codes
MSPILEUP_GENERIC_ERROR = 1
MSPILEUP_STORAGE_ERROR = 2


class MSTransferorError(WMException):
    """
    MSTransferorError represents generic MSTransferor error
    """
    def __init__(self, msg, errorNo=None, **data):
        """
        Constructor of MSTransferorError class
        """
        super().__init__(self, msg)
        self.data = data
        self.code = errorNo
        self.assign(msg)

    def error(self):
        """
        JSON representation of MSTransferorError

        :return: MSTransferorError representation
        """
        edict = {'data': self.data, 'message': self.msg, 'code': self.code, 'error': 'MSTransferorError'}
        return edict

    def __str__(self):
        """
        String representation of MSTransferorError

        :return: human readable MSTransferorError representation
        """
        return json.dumps(self.error(), indent=4)

    def assign(self, msg="", code=0):
        """
        Assign proper message and error codes

        :param msg: string
        :param code: int
        :return: None
        """
        if msg:
            self.msg = msg
        else:
            self.msg = 'Generic MSTransferor error'
        if code > 0:
            self.code = code
        else:
            self.code = MSPILEUP_GENERIC_ERROR


class MSTransferorGenericError(MSTransferorError):
    """
    Generic MSTransferor exception
    """
    def __init__(self, msg, errorNo=None, **data):
        super().__init__(msg, errorNo, **data)
        if not msg:
            msg = "generic error"
        self.assign(msg=msg, code=MSPILEUP_GENERIC_ERROR)


class MSTransferorStorageError(MSTransferorError):
    """
    Storage MSTransferor exception
    """
    def __init__(self, msg, errorNo=None, **data):
        super().__init__(msg, errorNo, **data)
        if not msg:
            msg = "storage error"
        self.assign(msg=msg, code=MSPILEUP_STORAGE_ERROR)



from builtins import str as newstr, bytes as newbytes

from WMCore.REST.Error import *
import math
import re
import numbers

from Utils.Utilities import decodeBytesToUnicodeConditional, encodeUnicodeToBytesConditional
from Utils.PythonVersion import PY3, PY2

def return_message(main_err, custom_err):
    if custom_err:
        return custom_err
    return main_err

def _arglist(argname, kwargs):
    val = kwargs.get(argname, None)
    if val == None:
        return []
    elif not isinstance(val, list):
        return [ val ]
    else:
        return val

def _check_rx(argname, val, custom_err = None):
    if not isinstance(val, (newstr, newbytes)):
        raise InvalidParameter(return_message("Incorrect '%s' parameter" % argname, custom_err))
    try:
        return re.compile(val)
    except:
        raise InvalidParameter(return_message("Incorrect '%s' parameter" % argname, custom_err))

def _check_str(argname, val, rx, custom_err = None):
    """
    This is not really check val is ASCII.
    2021 09: we are now using version 17.4.0 -> we do not need to convert to 
    bytes here anymore, we are using a recent verison of cherrypy.
    We merged the funcionality of _check_str and _check_ustr into a single function

    :type val: str or bytes (only utf8 encoded string) in py3, unicode or str in py2
    :type rx: regex, compiled from native str (unicode in py3, bytes in py2)
    """
    val = decodeBytesToUnicodeConditional(val, condition=PY3)
    val = encodeUnicodeToBytesConditional(val, condition=PY2)
    # `val` should now be a "native str" (unicode in py3, bytes in py2)
    # here str has not been redefined. it is default `str` in both py2 and py3.
    if not isinstance(val, str) or not rx.match(val):
        raise InvalidParameter(return_message("Incorrect '%s' parameter %s %s" % (argname, type(val), val), custom_err))
    return val

def _check_num(argname, val, bare, minval, maxval, custom_err = None):
    if not isinstance(val, numbers.Integral) and (not isinstance(val, (newstr, newbytes)) or (bare and not val.isdigit())):
        raise InvalidParameter(return_message("Incorrect '%s' parameter" % argname, custom_err))
    try:
        n = int(val)
        if (minval != None and n < minval) or (maxval != None and n > maxval):
            raise InvalidParameter(return_message("Parameter '%s' value out of bounds" % argname, custom_err))
        return n
    except InvalidParameter:
        raise
    except:
        raise InvalidParameter(return_message("Invalid '%s' parameter" % argname, custom_err))

def _check_real(argname, val, special, minval, maxval, custom_err = None):
    if not isinstance(val, numbers.Number) and not isinstance(val, (newstr, newbytes)):
        raise InvalidParameter(return_message("Incorrect '%s' parameter" % argname, custom_err))
    try:
        n = float(val)
        if not special and (math.isnan(n) or math.isinf(n)):
            raise InvalidParameter(return_message("Parameter '%s' improper value" % argname, custom_err))
        if (minval != None and n < minval) or (maxval != None and n > maxval):
            raise InvalidParameter(return_message("Parameter '%s' value out of bounds" % argname, custom_err))
        return n
    except InvalidParameter:
        raise
    except:
        raise InvalidParameter(return_message("Invalid '%s' parameter" % argname, custom_err))

def _validate_one(argname, param, safe, checker, optional, *args):
    val = param.kwargs.get(argname, None)
    if optional and val == None:
        safe.kwargs[argname] = None
    else:
        safe.kwargs[argname] = checker(argname, val, *args)
        del param.kwargs[argname]

def _validate_all(argname, param, safe, checker, *args):
    safe.kwargs[argname] = [checker(argname, v, *args) for v in _arglist(argname, param.kwargs)]
    if argname in param.kwargs:
        del param.kwargs[argname]

def validate_rx(argname, param, safe, optional = False, custom_err = None):
    """Validates that an argument is a valid regexp.

    Checks that an argument named `argname` exists in `param.kwargs`,
    and it a string which compiles into a python regular expression.
    If successful, the regexp object (not the string) is copied into
    `safe.kwargs` and the string value is removed from `param.kwargs`.

    If `optional` is True, the argument is not required to exist in
    `param.kwargs`; None is then inserted into `safe.kwargs`. Otherwise
    a missing value raises an exception."""
    _validate_one(argname, param, safe, _check_rx, optional, custom_err)

def validate_str(argname, param, safe, rx, optional = False, custom_err = None):
    """Validates that an argument is a string and matches a regexp.

    Checks that an argument named `argname` exists in `param.kwargs`
    and it is a string which matches regular expression `rx`. If
    successful the string is copied into `safe.kwargs` and the value
    is removed from `param.kwargs`.

    Accepts both unicode strings and utf8-encoded bytes strings as argument 
    string.
    Accepts regex compiled only with "native strings", which means  str in both 
    py2 and py3 (unicode in py3, bytes of utf8-encoded strings in py2)

    If `optional` is True, the argument is not required to exist in
    `param.kwargs`; None is then inserted into `safe.kwargs`. Otherwise
    a missing value raises an exception."""
    _validate_one(argname, param, safe, _check_str, optional, rx, custom_err)

def validate_num(argname, param, safe, optional = False,
                 bare = False, minval = None, maxval = None, custom_err = None):
    """Validates that an argument is a valid integer number.

    Checks that an argument named `argname` exists in `param.kwargs`,
    and it is an int or a string convertible to a valid number. If successful
    the integer value (not the string) is copied into `safe.kwargs`
    and the original int/string value is removed from `param.kwargs`.

    If `optional` is True, the argument is not required to exist in
    `param.kwargs`; None is then inserted into `safe.kwargs`. Otherwise
    a missing value raises an exception.

    If `bare` is True, the number is required to be a pure digit sequence if it is a string.
    Otherwise anything accepted by `int(val)` is acceted, including for
    example leading white space or sign. Note that either way arbitrarily
    large values are accepted; if you want to prevent abuse against big
    integers, use the `minval` and `maxval` thresholds described below,
    or check the length the of the string against some limit first.

    If `minval` or `maxval` are given, values less than or greater than,
    respectively, the threshold are rejected."""
    _validate_one(argname, param, safe, _check_num, optional, bare, minval, maxval, custom_err)

def validate_real(argname, param, safe, optional = False,
                  special = False, minval = None, maxval = None, custom_err = None):
    """Validates that an argument is a valid real number.

    Checks that an argument named `argname` exists in `param.kwargs`,
    and it is float number or a string convertible to a valid number. If successful
    the float value (not the string) is copied into `safe.kwargs`
    and the original float/string value is removed from `param.kwargs`.

    If `optional` is True, the argument is not required to exist in
    `param.kwargs`; None is then inserted into `safe.kwargs`. Otherwise
    a missing value raises an exception.

    Anything accepted by `float(val)` is accepted, including for example
    leading white space, sign and exponent. However NaN and +/- Inf are
    rejected unless `special` is True.

    If `minval` or `maxval` are given, values less than or greater than,
    respectively, the threshold are rejected."""
    _validate_one(argname, param, safe, _check_real, optional, special, minval, maxval, custom_err)

def validate_rxlist(argname, param, safe, custom_err = None):
    """Validates that an argument is an array of strings, each of which
    can be compiled into a python regexp object.

    Checks that an argument named `argname` is either a single string or
    an array of strings, each of which compiles into a regular expression.
    If successful the array is copied into `safe.kwargs` and the value is
    removed from `param.kwargs`. The value always becomes an array in
    `safe.kwargs`, even if no or only one argument was provided.

    Note that an array of zero length is accepted, meaning there were no
    `argname` parameters at all in `param.kwargs`."""
    _validate_all(argname, param, safe, _check_rx, custom_err)

def validate_strlist(argname, param, safe, rx, custom_err = None):
    """Validates that an argument is an array of strings, each of which
    matches a regexp.

    Checks that an argument named `argname` is either a single string or
    an array of strings, each of which matches the regular expression
    `rx`. If successful the array is copied into `safe.kwargs` and the
    value is removed from `param.kwargs`. The value always becomes an
    array in `safe.kwargs`, even if no or only one argument was provided.

    Use this method only for inputs which are meant to be bare strings.

    Note that an array of zero length is accepted, meaning there were no
    `argname` parameters at all in `param.kwargs`."""
    _validate_all(argname, param, safe, _check_str, rx, custom_err)

def validate_numlist(argname, param, safe, bare=False, minval=None, maxval=None, custom_err = None):
    """Validates that an argument is an array of integers, as checked by
    `validate_num()`.

    Checks that an argument named `argname` is either a single string/int or
    an array of strings/int, each of which validates with `validate_num` and
    `bare`, `minval` and `maxval` arguments.  If successful the array is
    copied into `safe.kwargs` and the value is removed from `param.kwargs`.
    The value always becomes an array in `kwsafe`, even if no or only one
    argument was provided.

    Note that an array of zero length is accepted, meaning there were no
    `argname` parameters at all in `param.kwargs`."""
    _validate_all(argname, param, safe, _check_num, bare, minval, maxval, custom_err)

def validate_reallist(argname, param, safe, special=False, minval=None, maxval=None, custom_err = None):
    """Validates that an argument is an array of integers, as checked by
    `validate_real()`.

    Checks that an argument named `argname` is either a single string/float or
    an array of strings/floats, each of which validates with `validate_real` and
    `special`, `minval` and `maxval` arguments.  If successful the array is
    copied into `safe.kwargs` and the value is removed from `param.kwargs`.
    The value always becomes an array in `safe.kwargs`, even if no or only
    one argument was provided.

    Note that an array of zero length is accepted, meaning there were no
    `argname` parameters at all in `param.kwargs`."""
    _validate_all(argname, param, safe, _check_real, special, minval, maxval, custom_err)

def validate_no_more_input(param):
    """Verifies no more input is left in `param.args` or `param.kwargs`."""
    if param.args:
        raise InvalidParameter("Excess path arguments, not validated args='%s'" % param.args)
    if param.kwargs:
        raise InvalidParameter("Excess keyword arguments, not validated kwargs='%s'" % param.kwargs)

def validate_lengths(safe, *names):
    """Verifies that all `names` exist in `safe.kwargs`, are lists, and
    all the lists have the same length. This is convenience function for
    checking that an API accepting multiple values receives equal number
    of values for all of its parameters."""
    refname = names[0]
    if refname not in safe.kwargs or not isinstance(safe.kwargs[refname], list):
        raise InvalidParameter("Incorrect '%s' parameter" % refname)

    reflen = len(safe.kwargs[refname])
    for other in names[1:]:
        if other not in safe.kwargs or not isinstance(safe.kwargs[other], list):
            raise InvalidParameter("Incorrect '%s' parameter" % other)
        elif len(safe.kwargs[other]) != reflen:
            raise InvalidParameter("Mismatched number of arguments: %d %s vs. %d %s"
                                   % (reflen, refname, len(safe.kwargs[other]), other))
